import path from "path";
import * as brokenSemver from "semver";
import { get, last } from "bottom-line-utils";
import fs from "fs-extra";
import { jsxMigration, migrationsConfig } from "./migrations.config.js";
import { Migration } from "./Migration.js";
import Question from "./Question.js";
const semver = {
    gte: (brokenSemver.gte || brokenSemver.default.gte),
};
const applyMigrations = async ({ migration, migrations, updateVersion }) => {
    let skipped = 0;
    const err = (e) => {
        skipped++;
        console.warn("✖", e.message);
    };
    for (let i = 0; i < migrations.length; i++) {
        const migrationConfig = migrations[i];
        console.info(`Upgrading [${migrationConfig.version} -> ${migrationConfig.nextVersion}]`);
        for (let j = 0; j < migrationConfig.steps.length; j++) {
            const step = migrationConfig.steps[j];
            if ("jsx" in step && step.jsx !== migration.jsx) {
                continue;
            }
            console.info();
            console.info(`- ${step.name}`);
            await step.fn(migration).then(() => { console.warn("✔ ok"); }, err);
        }
    }
    const migCfg = last(migrations);
    if (migCfg && updateVersion) {
        await migration.setPath("libraryTemplate.version", migCfg.nextVersion);
    }
    return skipped;
};
const aggressiveMessage = async (migrations) => {
    const aggressive = migrations.map(m => m.aggresive).filter(Boolean);
    if (!aggressive.length) {
        return true;
    }
    console.info("");
    console.info("WARNING: Aggressive upgrade detected!");
    console.info("Aggressive upgrade means that customizations applied over previously-generated library may be lost "
        + "or that library won't work after upgrading without further changes to the code. Please verify this list "
        + "of POTENTIAL POSSIBLE issues and accept the upgrade progress:");
    console.info(aggressive.map(a => "- " + a.trim()).join("\n"));
    console.info("");
    const q = new Question();
    const accept = await q.ask("Do you want to continue? (y/n) [n]");
    q.close();
    return accept.startsWith("y");
};
const prepareMigrate = async (targetDir) => {
    let pkg, lang, ver, rethrow = false;
    const pkgPath = path.join(targetDir, "package.json");
    try {
        pkg = JSON.parse(String(await fs.readFile(pkgPath)));
        ver = get(pkg, ["libraryTemplate", "version"]);
        if (!ver) {
            throw new Error("No version");
        }
        lang = get(pkg, ["libraryTemplate", "language"]);
        if (lang !== "typescript") {
            rethrow = true;
            throw new Error("Migration from js-library-template is currently not supported");
        }
    }
    catch (e) {
        if (rethrow) {
            throw e;
        }
        throw new Error("Target directory is not empty, no supported library found to upgrade.");
    }
    return {
        pkg, ver,
    };
};
const migrate = async (targetDir) => {
    const { pkg, ver } = await prepareMigrate(targetDir);
    const migration = new Migration({ targetDir, pkg });
    const migrations = migrationsConfig.filter(migrationConfig => {
        return semver.gte(migrationConfig.version, ver);
    });
    const versions = migrations.map(m => m.nextVersion);
    const cont = await aggressiveMessage(migrations);
    if (!cont) {
        console.info("Upgrade stopped.");
        return;
    }
    if (!migrations.length) {
        console.info("The project is up to date with current ts-library-template.");
        return;
    }
    console.info(`${migrations.length} updates to apply, versions: ${versions.join(", ")}`);
    const updateVersion = true;
    const skipped = await applyMigrations({ migration, migrations, updateVersion });
    console.info();
    if (skipped) {
        console.info(`Skipped ${skipped} upgrades. Please perform them manually if needed.`);
    }
    console.info("Upgrading finished.");
};
const migrateJsx = async (targetDir) => {
    const { pkg } = await prepareMigrate(targetDir);
    if (pkg.libraryTemplate.jsx) {
        return;
    }
    const migration = new Migration({ targetDir, pkg });
    const migrations = [jsxMigration];
    const updateVersion = false;
    const skipped = await applyMigrations({ migration, migrations, updateVersion });
    console.info();
    if (skipped) {
        console.info(`Skipped ${skipped} upgrades. Please perform them manually if needed.`);
    }
    console.info("Upgrading finished.");
};
export { migrate, migrateJsx };
//# sourceMappingURL=migrate.js.map